/*
                Range Generator
                ===============

    Copyright (c) 2009-2011 Khaled Alshaya

    Distributed under the Boost Software License, version 1.0
    (See the license at: http://www.boost.org/license_1_0.txt).
*/

/*
                Range Test
                ==========

    In this code, you will find the test cases that
    are written to test Range. The test cases cover
    as much as possible (I hope so :).
*/

#define CATCH_CONFIG_MAIN
#include "catch.hpp"
#include "../range.hpp"
using namespace Range;

TEST_CASE("range/constructor", "Default constructor should construct a valid empty range."){
    REQUIRE_NOTHROW( range() );
    REQUIRE( range().empty() );
}
TEST_CASE("range/constructor", "An exception should be thrown in case of invalid range."){
    REQUIRE_THROWS( range(1, 10, 0) );
    REQUIRE_THROWS( range(0, 10, -1) );
    REQUIRE_THROWS( range(10, 0, 1) );
    REQUIRE_NOTHROW( range(10, 10, 1) );
    REQUIRE_NOTHROW( range(10, 10, -1) );
    REQUIRE_NOTHROW( range(0, 10); )
    REQUIRE_NOTHROW( range(10, 0); )
    REQUIRE( range(0, 10).size() == 10 );
    REQUIRE( range(10, 0).size() == 10 );
    REQUIRE( range(10).size() == 10 );
    REQUIRE( range(-10).size() == 10 );
}
TEST_CASE("range/operator", "An assigned or copy constructed range should be equal to the original range."){
    range r1(0, 10, 1), r2;
    r2 = r1;
    REQUIRE(r1 == r2);
    r1 = range(1, 10, 1);
    REQUIRE_FALSE(r1 == r2);
    REQUIRE(r1 != r2);
}
TEST_CASE("range/relational_operator", "Relational comparison is a lexicographical comparison."){
    REQUIRE( range(0, 10, 1) < range(1, 11, 1) );
    REQUIRE( range(0, 10, 1) <= range(0, 10, 1) );
    REQUIRE( range(1, 11, 1) > range(0, 10, 1) );
    REQUIRE( range(0, 10, 1) >= range(0, 10, 1) );
}
TEST_CASE("range/size", "The size() of a range should equal its max_size() because range is immutable container."){
    range r(0, 10, 2);
    REQUIRE( r.size() == 5 );
    REQUIRE( r.size() == r.max_size() );
    REQUIRE( !r.empty() )
}
TEST_CASE("range/bound", "An iterator created manually should equal to the iterators created by range member functions."){
    range r(0, 10, 2);
    REQUIRE( r.begin() == range::iterator(&r, 0) );
    REQUIRE( r.end() == range::iterator(&r, 5) );
    REQUIRE( r.rbegin() == range::reverse_iterator(&r, 0) );
    REQUIRE( r.rend() == range::reverse_iterator(&r, 5) );
}
TEST_CASE("range/iterator", "Testing the iterator operators."){
    typedef range::iterator iter;
    range r(0, 10, 2);
    iter i(&r, 4), b(&r, 0), e(&r, 4);
    REQUIRE( i == iter(&r, 4) );
    i = iter(&r, 3);
    REQUIRE( i == iter(&r, 3) );
    REQUIRE( i != iter(&r, 0) );
    REQUIRE( i < iter(&r, 4) );
    REQUIRE( i <= iter(&r, 3) );
    REQUIRE( i > iter(&r, 2) );
    REQUIRE( i >= iter(&r, 3) );
    REQUIRE( *i == 6 );
    ++i;
    REQUIRE( i == iter(&r, 4) );
    --i;
    REQUIRE( i == iter(&r, 3) );
    i++;
    REQUIRE( i == iter(&r, 4) );
    i--;
    REQUIRE( i == iter(&r, 3) );
    REQUIRE( (e-b) == 4 );
    REQUIRE( (b+4) == e );
    REQUIRE( (e-4) == b );
    b += 4;
    REQUIRE( b == e );
    b -= 4;
    e -= 4;
    REQUIRE( b == e );
    REQUIRE( *b == *e );
    REQUIRE( *b == b[0] );
    e += 4;
    REQUIRE( *e == b[e-b] );
}
TEST_CASE("range/reverse_iterator", "Testing the reverse_iterator operators."){
    typedef range::reverse_iterator iter;
    range r(0, 10, 2);
    iter i(&r, 4), b(&r, 0), e(&r, 4);
    REQUIRE( i == iter(&r, 4) );
    i = iter(&r, 3);
    REQUIRE( i == iter(&r, 3) );
    REQUIRE( i != iter(&r, 0) );
    REQUIRE( i < iter(&r, 4) );
    REQUIRE( i <= iter(&r, 3) );
    REQUIRE( i > iter(&r, 2) );
    REQUIRE( i >= iter(&r, 3) );
    REQUIRE( *i == 2 );
    ++i;
    REQUIRE( i == iter(&r, 4) );
    --i;
    REQUIRE( i == iter(&r, 3) );
    i++;
    REQUIRE( i == iter(&r, 4) );
    i--;
    REQUIRE( i == iter(&r, 3) );
    REQUIRE( (e-b) == 4 );
    REQUIRE( (b+4) == e );
    REQUIRE( (e-4) == b );
    b += 4;
    REQUIRE( b == e );
    b -= 4;
    e -= 4;
    REQUIRE( b == e );
    REQUIRE( *b == *e );
    REQUIRE( *b == b[0] );
    e += 4;
    REQUIRE( *e == b[e-b] );
}
TEST_CASE("range/exist", "Given an element, the function should determine if it belones to the range."){
    range r(1, 11, 3);
    REQUIRE( !r.exist(0) );
    REQUIRE( r.exist(1) );
    REQUIRE( !r.exist(2) );
    REQUIRE( !r.exist(3) );
    REQUIRE( r.exist(4) );
    REQUIRE( !r.exist(5) );
    REQUIRE( !r.exist(6) );
    REQUIRE( r.exist(7) );
    REQUIRE( !r.exist(8) );
    REQUIRE( !r.exist(9) );
    REQUIRE( r.exist(10) );
    REQUIRE( !r.exist(11) );
    r = range(11, 1, -3);
    REQUIRE( !r.exist(12) );
    REQUIRE( r.exist(11) );
    REQUIRE( !r.exist(10) );
    REQUIRE( !r.exist(9) );
    REQUIRE( r.exist(8) );
    REQUIRE( !r.exist(7) );
    REQUIRE( !r.exist(6) );
    REQUIRE( r.exist(5) );
    REQUIRE( !r.exist(4) );
    REQUIRE( !r.exist(3) );
    REQUIRE( r.exist(2) );
    REQUIRE( !r.exist(1) );
}
TEST_CASE("range/find", "Given an element, the function should return an iterator to its position."){
    range r(1, 11, 3);
    REQUIRE( r.find(0) == r.end() );
    REQUIRE( r.find(1) == (r.begin()+0) );
    REQUIRE( r.find(2) == r.end() );
    REQUIRE( r.find(3) == r.end() );
    REQUIRE( r.find(4) == (r.begin()+1) );
    REQUIRE( r.find(5) == r.end() );
    REQUIRE( r.find(6) == r.end() );
    REQUIRE( r.find(7) == (r.begin()+2) );
    REQUIRE( r.find(8) == r.end() );
    REQUIRE( r.find(9) == r.end() );
    REQUIRE( r.find(10) == (r.begin()+3) );
    REQUIRE( r.find(11) == r.end() );
    r = range(11, 1, -3);
    REQUIRE( r.find(12) == r.end() );
    REQUIRE( r.find(11) == (r.begin()+0) );
    REQUIRE( r.find(10) == r.end() );
    REQUIRE( r.find(9) == r.end());
    REQUIRE( r.find(8) == (r.begin()+1) );
    REQUIRE( r.find(7) == r.end() );
    REQUIRE( r.find(6) == r.end() );
    REQUIRE( r.find(5) == (r.begin()+2) );
    REQUIRE( r.find(4) == r.end() );
    REQUIRE( r.find(3) == r.end() );
    REQUIRE( r.find(2) == (r.begin()+3) );
    REQUIRE( r.find(1) == r.end() );
}